import functools
from collections.abc import Callable, Coroutine, Hashable, Sequence
from enum import Enum
from inspect import signature
from typing import (
    Any,
    Concatenate,
    Generic,
    Literal,
    ParamSpec,
    TypeVar,
    cast,
    get_args,
    get_origin,
    overload,
)

from typing_extensions import Self

from gyver.misc.casting import as_async

T = TypeVar('T')
P = ParamSpec('P')
SelfT = TypeVar('SelfT')


def cache(f: Callable[P, T]) -> Callable[P, T]:
    return cast(Callable[P, T], functools.cache(f))


class MethodType(Enum):
    SELF_ONLY = 0
    HASHABLE_ARGS = 1
    UNKNOWN_OR_UNHASHABLE = 2


class lazymethod(Generic[SelfT, T, P]):
    __slots__ = ('_func', 'public_name', 'private_name', '_method_type', '_container')

    format_ = '_lazymethod_{method_name}_'

    def __init__(self, func: Callable[Concatenate[SelfT, P], T]) -> None:
        self._func = func
        self._method_type = self._determine_method_type(func)
        self._container = self._get_container()

    def _determine_method_type(self, func: Callable) -> MethodType:
        sig = signature(func)
        if len(sig.parameters) == 1:
            return MethodType.SELF_ONLY
        elif all(
            param.annotation is not param.empty and self._is_hashable(param.annotation)
            for param in tuple(sig.parameters.values())[1:]
        ):
            return MethodType.HASHABLE_ARGS
        else:
            return MethodType.UNKNOWN_OR_UNHASHABLE

    def _is_hashable(self, annotation: Any) -> bool:
        if origin := get_origin(annotation):
            args = get_args(annotation)
            return self._is_hashable(origin) and all(
                self._is_hashable(arg) for arg in args
            )
        if isinstance(annotation, type):
            return issubclass(annotation, Hashable)
        return False

    def _get_container(self) -> type[dict[Hashable, T] | list[tuple[Any, T]]] | None:
        match self._method_type:
            case MethodType.SELF_ONLY:
                return None
            case MethodType.HASHABLE_ARGS:
                return dict
            case MethodType.UNKNOWN_OR_UNHASHABLE:
                return list

    def __set_name__(self, owner: type[SelfT], name: str) -> None:
        self.public_name = name
        self.private_name = self.format_.format(method_name=name)

    @classmethod
    def get_private(cls, name: str) -> str:
        return cls.format_.format(method_name=name)

    @overload
    def __get__(self, instance: None, owner: type[SelfT]) -> Self: ...

    @overload
    def __get__(self, instance: SelfT, owner: type[SelfT]) -> Callable[P, T]: ...

    def __get__(
        self, instance: SelfT | None, owner: type[SelfT]
    ) -> Callable[P, T] | Self:
        if instance is None:
            return self
        return self._call(instance)

    def _call(self, instance: SelfT) -> Callable[P, T]:
        @functools.wraps(self._func)
        def _callable(*args: P.args, **kwargs: P.kwargs) -> T:
            if (
                value := self._get(
                    instance,
                    self.private_name,
                    args,
                    kwargs,
                )
            ) is not None:
                return value
            return self._set(instance, *args, **kwargs)

        return _callable

    def _set(self, instance: SelfT, *args: P.args, **kwargs: P.kwargs) -> T:
        value = self._func(instance, *args, **kwargs)
        if self._container is None:
            object.__setattr__(instance, self.private_name, value)
        elif self._container is dict:
            container = cast(
                dict[Hashable, T],
                getattr(instance, self.private_name, self._container()),
            )
            container[(*args, tuple(kwargs.items()))] = value
            object.__setattr__(instance, self.private_name, container)
        elif self._container is list:
            container = cast(
                list[tuple[Any, T]],
                getattr(instance, self.private_name, self._container()),
            )
            container.append(((*args, tuple(kwargs.items())), value))
            object.__setattr__(instance, self.private_name, container)
        return value

    def _get(
        self, instance: SelfT, name: str, args: Sequence[Any], kwargs: dict[str, Any]
    ) -> T | None:
        if self._container is None:
            return getattr(instance, name, None)
        elif self._container is dict:
            container = getattr(instance, name, None)
            if container is None:
                return None
            return container.get((*args, tuple(kwargs.items())))
        elif self._container is list:
            container = getattr(instance, name, None)
            if container is None:
                return None
            for item in container:
                if item[0] == (*args, tuple(kwargs.items())):
                    return item[1]
            return None

    @classmethod
    def is_initialized(cls, instance: SelfT, name: str) -> bool:
        return hasattr(instance, cls.format_.format(method_name=name))


@overload
def make_noop(
    *, returns: T = None, asyncio: Literal[False] = False
) -> Callable[..., T]: ...


@overload
def make_noop(
    *, returns: T = None, asyncio: Literal[True]
) -> Callable[..., Coroutine[Any, Any, T]]: ...


def make_noop(
    *, returns: T = None, asyncio: bool = False
) -> Callable[..., T] | Callable[..., Coroutine[Any, Any, T]]:
    def _noop(*args, **kwargs) -> T:
        return returns

    return _noop if not asyncio else as_async(_noop)



from collections.abc import Generator
from contextlib import contextmanager
from typing import Any


@contextmanager
def raise_insteadof(
    insteadof: type[Exception], exc: type[Exception] = Exception, *args: Any
) -> Generator[None, None, None]:
    try:
        yield
    except insteadof:
        raise exc(*args) from None


class Namespace:
    """
    A lightweight class that provides dictionary-like functionality but allows
    access to its keys as if they were attributes using dot notation.

    The `Namespace` is a thin wrapper around a dictionary, offering attribute-style
    access (`namespace.key`) in addition to the traditional item-style access
    (`namespace['key']`). It is designed for cases where dynamic attribute management
    is needed while maintaining the flexibility and mutability of a standard dictionary.

    Attributes:
        __namespace_data__ (dict): The underlying dictionary used for storage of keys
        and values.

    Methods:
        __getattr__(name): Returns the value associated with `name` if it exists.
        Raises an AttributeError if not found.

        __setattr__(name, value): Sets the value of `name` to `value`. Respects the
        `__slots__` definition for internal attributes.

        __getitem__(name): Retrieves the value associated with `name` using dictionary
        indexing.

        __setitem__(name, value): Sets the value associated with `name` using dictionary
        indexing.

        __delitem__(name): Deletes the item associated with `name` from the underlying
        dictionary.

        __delattr__(name): Deletes the attribute-style `name` from the underlying dictionary.

        __iter__(): Iterates over the keys of the namespace.

        __len__(): Returns the number of items in the namespace.

        get(name, default=None): Retrieves the value associated with `name`, or returns
        `default` if the key is not found.

        __repr__(): Returns a string representation of the underlying dictionary.
    """

    __slots__ = ('__namespace_data__',)

    def __init__(self, initial: dict[str, Any] | None = None) -> None:
        self.__namespace_data__ = initial or {}

    def __getattr__(self, name: str) -> Any:
        with raise_insteadof(
            KeyError, AttributeError, f'Attribute {name} not found in namespace.'
        ):
            return self.__namespace_data__[name]

    def __setattr__(self, name: str, value: Any, /) -> None:
        if name in Namespace.__slots__:
            return super().__setattr__(name, value)
        self.__namespace_data__[name] = value

    def __getitem__(self, name: str) -> Any:
        return self.__namespace_data__[name]

    def __setitem__(self, name: str, value: Any, /) -> None:
        self.__namespace_data__[name] = value

    def __delitem__(self, name: str, /) -> None:
        del self.__namespace_data__[name]

    def __delattr__(self, name: str, /) -> None:
        if name in Namespace.__slots__:
            return super().__delattr__(name)

        with raise_insteadof(
            KeyError, AttributeError, f'Attribute {name} not found in namespace.'
        ):
            del self.__namespace_data__[name]

    def __iter__(self) -> Any:
        return iter(self.__namespace_data__)

    def __len__(self) -> int:
        return len(self.__namespace_data__)

    def get(self, name: str, default: Any = None, /) -> Any:
        try:
            return self.__namespace_data__[name]
        except KeyError:
            return default

    def __repr__(self) -> str:
        return repr(self.__namespace_data__)



from typing import TypeVar

from gyver.misc.strings import exclamation, question
from gyver.misc.strings import sentence as _sentence

ExceptionT = TypeVar('ExceptionT', bound=Exception)


def sentence(exc_type: type[ExceptionT], message: str, *args) -> ExceptionT:
    """Add a period to the end of the message."""
    return exc_type(_sentence(message), *args)


def scream(exc_type: type[ExceptionT], message: str, *args) -> ExceptionT:
    """Add an exclamation point to the end of the message."""
    return exc_type(exclamation(message), *args)


def question_message(exc_type: type[ExceptionT], message: str, *args) -> ExceptionT:
    """Add a question mark to the end of the message."""
    return exc_type(question(message), *args)
